# Copyright 2015 EMC Corporation
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Isilon specific NAS backend plugin.
"""
import os
from oslo_config import cfg
from oslo_log import log
from oslo_utils import units
from requests.exceptions import HTTPError
import six

from manila.common import constants as const
from manila import exception
from manila.i18n import _
from manila.share.drivers.dell_emc.plugins import base
from manila.share.drivers.dell_emc.plugins.isilon import isilon_api

CONF = cfg.CONF
VERSION = "0.1.0"

LOG = log.getLogger(__name__)


class IsilonStorageConnection(base.StorageConnection):
    """Implements Isilon specific functionality for EMC Manila driver."""

    def __init__(self, *args, **kwargs):
        super(IsilonStorageConnection, self).__init__(*args, **kwargs)
        self._server = None
        self._port = None
        self._username = None
        self._password = None
        self._server_url = None
        self._connect_resp = None
        self._root_dir = None
        self._verify_ssl_cert = None
        self._containers = {}
        self._shares = {}
        self._snapshots = {}

        self._isilon_api = None
        self._isilon_api_class = isilon_api.IsilonApi
        self.driver_handles_share_servers = False
        self.dhss_mandatory_security_service_association = {}

    def _get_container_path(self, share):
        """Return path to a container."""
        return os.path.join(self._root_dir, share['name'])

    def create_share(self, context, share, share_server):
        """Is called to create share."""
        if share['share_proto'] == 'NFS':
            location = self._create_nfs_share(share)
        elif share['share_proto'] == 'CIFS':
            location = self._create_cifs_share(share)
        else:
            message = (_('Unsupported share protocol: %(proto)s.') %
                       {'proto': share['share_proto']})
            LOG.error(message)
            raise exception.InvalidShare(reason=message)

        # apply directory quota based on share size
        max_share_size = share['size'] * units.Gi
        self._isilon_api.quota_create(
            self._get_container_path(share), 'directory', max_share_size)

        return location

    def create_share_from_snapshot(self, context, share, snapshot,
                                   share_server):
        """Creates a share from the snapshot."""

        # Create share at new location
        location = self.create_share(context, share, share_server)

        # Clone snapshot to new location
        fq_target_dir = self._get_container_path(share)
        self._isilon_api.clone_snapshot(snapshot['name'], fq_target_dir)

        return location

    def _create_nfs_share(self, share):
        """Is called to create nfs share."""
        container_path = self._get_container_path(share)
        self._isilon_api.create_directory(container_path)

        share_created = self._isilon_api.create_nfs_export(container_path)
        if not share_created:
            message = (
                _('The requested NFS share "%(share)s" was not created.') %
                {'share': share['name']})
            LOG.error(message)
            raise exception.ShareBackendException(msg=message)
        location = '{0}:{1}'.format(self._server, container_path)
        return location

    def _create_cifs_share(self, share):
        """Is called to create cifs share."""

        # Create the directory
        container_path = self._get_container_path(share)
        self._isilon_api.create_directory(container_path)
        self._isilon_api.create_smb_share(share['name'], container_path)
        share_path = '\\\\{0}\\{1}'.format(self._server, share['name'])
        return share_path

    def create_snapshot(self, context, snapshot, share_server):
        """Is called to create snapshot."""
        snapshot_path = os.path.join(self._root_dir, snapshot['share_name'])
        self._isilon_api.create_snapshot(snapshot['name'], snapshot_path)

    def delete_share(self, context, share, share_server):
        """Is called to remove share."""
        if share['share_proto'] == 'NFS':
            self._delete_nfs_share(share)
        elif share['share_proto'] == 'CIFS':
            self._delete_cifs_share(share)
        else:
            message = (_('Unsupported share type: %(type)s.') %
                       {'type': share['share_proto']})
            LOG.error(message)
            raise exception.InvalidShare(reason=message)

    def _delete_nfs_share(self, share):
        """Is called to remove nfs share."""
        share_id = self._isilon_api.lookup_nfs_export(
            self._root_dir + '/' + share['name'])

        if share_id is None:
            lw = ('Attempted to delete NFS Share "%s", but the share does '
                  'not appear to exist.')
            LOG.warning(lw, share['name'])
        else:
            # attempt to delete the share
            export_deleted = self._isilon_api.delete_nfs_share(share_id)
            if not export_deleted:
                message = _('Error deleting NFS share: %s') % share['name']
                LOG.error(message)
                raise exception.ShareBackendException(msg=message)

    def _delete_cifs_share(self, share):
        """Is called to remove CIFS share."""
        smb_share = self._isilon_api.lookup_smb_share(share['name'])
        if smb_share is None:
            lw = ('Attempted to delete CIFS Share "%s", but the share does '
                  'not appear to exist.')
            LOG.warning(lw, share['name'])
        else:
            share_deleted = self._isilon_api.delete_smb_share(share['name'])
            if not share_deleted:
                message = _('Error deleting CIFS share: %s') % share['name']
                LOG.error(message)
                raise exception.ShareBackendException(msg=message)

    def delete_snapshot(self, context, snapshot, share_server):
        """Is called to remove snapshot."""
        self._isilon_api.delete_snapshot(snapshot['name'])

    def ensure_share(self, context, share, share_server):
        """Invoked to ensure that share is exported."""

    def extend_share(self, share, new_size, share_server=None):
        """Extends a share."""
        new_quota_size = new_size * units.Gi
        self._isilon_api.quota_set(
            self._get_container_path(share), 'directory', new_quota_size)

    def allow_access(self, context, share, access, share_server):
        """Allow access to the share."""

        if share['share_proto'] == 'NFS':
            self._nfs_allow_access(share, access)
        elif share['share_proto'] == 'CIFS':
            self._cifs_allow_access(share, access)
        else:
            message = _(
                'Unsupported share protocol: %s. Only "NFS" and '
                '"CIFS" are currently supported share protocols.') % share[
                'share_proto']
            LOG.error(message)
            raise exception.InvalidShare(reason=message)

    def _nfs_allow_access(self, share, access):
        """Allow access to nfs share."""
        access_type = access['access_type']
        if access_type != 'ip':
            message = _('Only "ip" access type allowed for the NFS '
                        'protocol.')
            LOG.error(message)
            raise exception.InvalidShareAccess(reason=message)

        export_path = self._get_container_path(share)
        access_ip = access['access_to']
        access_level = access['access_level']
        share_id = self._isilon_api.lookup_nfs_export(export_path)

        share_access_group = 'clients'
        if access_level == const.ACCESS_LEVEL_RO:
            share_access_group = 'read_only_clients'

        # Get current allowed clients
        export = self._get_existing_nfs_export(share_id)
        current_clients = export[share_access_group]

        # Format of ips could be '10.0.0.2', or '10.0.0.2, 10.0.0.0/24'
        ips = list()
        ips.append(access_ip)
        ips.extend(current_clients)
        export_params = {share_access_group: ips}
        url = '{0}/platform/1/protocols/nfs/exports/{1}'.format(
            self._server_url, share_id)
        resp = self._isilon_api.request('PUT', url, data=export_params)
        resp.raise_for_status()

    def _cifs_allow_access(self, share, access):
        access_type = access['access_type']
        access_to = access['access_to']
        access_level = access['access_level']
        if access_type == 'ip':
            access_ip = access['access_to']
            self._cifs_allow_access_ip(access_ip, share, access_level)
        elif access_type == 'user':
            self._cifs_allow_access_user(access_to, share, access_level)
        else:
            message = _('Only "ip" and "user" access types allowed for '
                        'CIFS protocol.')
            LOG.error(message)
            raise exception.InvalidShareAccess(reason=message)

    def _cifs_allow_access_ip(self, ip, share, access_level):
        if access_level == const.ACCESS_LEVEL_RO:
            message = _('Only RW Access allowed for CIFS Protocol when using '
                        'the "ip" access type.')
            LOG.error(message)
            raise exception.InvalidShareAccess(reason=message)

        allowed_ip = 'allow:' + ip
        smb_share = self._isilon_api.lookup_smb_share(share['name'])
        host_acl = smb_share['host_acl']
        if allowed_ip not in host_acl:
            host_acl.append(allowed_ip)
            data = {'host_acl': host_acl}
            url = ('{0}/platform/1/protocols/smb/shares/{1}'
                   .format(self._server_url, share['name']))
            r = self._isilon_api.request('PUT', url, data=data)
            r.raise_for_status()

    def _cifs_allow_access_user(self, user, share, access_level):
        if access_level == const.ACCESS_LEVEL_RW:
            smb_permission = isilon_api.SmbPermission.rw
        elif access_level == const.ACCESS_LEVEL_RO:
            smb_permission = isilon_api.SmbPermission.ro
        else:
            message = _('Only "RW" and "RO" access levels are supported.')
            LOG.error(message)
            raise exception.InvalidShareAccess(reason=message)

        self._isilon_api.smb_permissions_add(share['name'], user,
                                             smb_permission)

    def deny_access(self, context, share, access, share_server):
        """Deny access to the share."""

        if share['share_proto'] == 'NFS':
            self._nfs_deny_access(share, access)
        elif share['share_proto'] == 'CIFS':
            self._cifs_deny_access(share, access)

    def _nfs_deny_access(self, share, access):
        """Deny access to nfs share."""
        if access['access_type'] != 'ip':
            return

        denied_ip = access['access_to']
        access_level = access['access_level']
        share_access_group = 'clients'
        if access_level == const.ACCESS_LEVEL_RO:
            share_access_group = 'read_only_clients'

        # Get list of currently allowed client ips
        export_id = self._isilon_api.lookup_nfs_export(
            self._get_container_path(share))
        if export_id is None:
            message = _('Share %s should have been created, but was not '
                        'found.') % share['name']
            LOG.error(message)
            raise exception.ShareBackendException(msg=message)
        export = self._get_existing_nfs_export(export_id)
        try:
            clients = export[share_access_group]
        except KeyError:
            message = (_('Export %(export_name)s should have contained the '
                         'JSON key %(json_key)s, but this key was not found.')
                       % {'export_name': share['name'],
                          'json_key': share_access_group})
            LOG.error(message)
            raise exception.ShareBackendException(msg=message)
        allowed_ips = set(clients)

        if allowed_ips.__contains__(denied_ip):
            allowed_ips.remove(denied_ip)
            data = {share_access_group: list(allowed_ips)}
            url = ('{0}/platform/1/protocols/nfs/exports/{1}'
                   .format(self._server_url, six.text_type(export_id)))
            r = self._isilon_api.request('PUT', url, data=data)
            r.raise_for_status()

    def _get_existing_nfs_export(self, export_id):
        export = self._isilon_api.get_nfs_export(export_id)
        if export is None:
            message = _('NFS share with export id %d should have been '
                        'created, but was not found.') % export_id
            LOG.error(message)
            raise exception.ShareBackendException(msg=message)

        return export

    def _cifs_deny_access(self, share, access):
        access_type = access['access_type']
        if access_type == 'ip':
            self._cifs_deny_access_ip(access['access_to'], share)
        elif access_type == 'user':
            self._cifs_deny_access_user(share, access)
        else:
            message = _('Access type for CIFS deny access request was '
                        '"%(access_type)s". Only "user" and "ip" access types '
                        'are supported for CIFS protocol access.') % {
                'access_type': access_type}
            LOG.warning(message)

    def _cifs_deny_access_ip(self, denied_ip, share):
        """Deny access to cifs share."""

        share_json = self._isilon_api.lookup_smb_share(share['name'])
        host_acl_list = share_json['host_acl']
        allow_ip = 'allow:' + denied_ip
        if allow_ip in host_acl_list:
            host_acl_list.remove(allow_ip)
            share_params = {"host_acl": host_acl_list}
            url = ('{0}/platform/1/protocols/smb/shares/{1}'
                   .format(self._server_url, share['name']))
            resp = self._isilon_api.request('PUT', url, data=share_params)
            resp.raise_for_status()

    def _cifs_deny_access_user(self, share, access):
        self._isilon_api.smb_permissions_remove(share['name'], access[
            'access_to'])

    def check_for_setup_error(self):
        """Check for setup error."""

    def connect(self, emc_share_driver, context):
        """Connect to an Isilon cluster."""
        self._server = emc_share_driver.configuration.safe_get(
            "emc_nas_server")
        self._port = (
            int(emc_share_driver.configuration.safe_get("emc_nas_server_port"))
        )
        self._server_url = ('https://' + self._server + ':' +
                            six.text_type(self._port))
        self._username = emc_share_driver.configuration.safe_get(
            "emc_nas_login")
        self._password = emc_share_driver.configuration.safe_get(
            "emc_nas_password")
        self._root_dir = emc_share_driver.configuration.safe_get(
            "emc_nas_root_dir")
        # TODO(Shaun Edwards): make verify ssl a config variable?
        self._verify_ssl_cert = False
        self._isilon_api = self._isilon_api_class(self._server_url, auth=(
            self._username, self._password),
            verify_ssl_cert=self._verify_ssl_cert)
        if not self._isilon_api.is_path_existent(self._root_dir):
            self._isilon_api.create_directory(self._root_dir, recursive=True)

    def update_share_stats(self, stats_dict):
        """TODO."""
        # TODO(Shaun Edwards): query capacity, set storage_protocol,
        # QoS support?
        stats_dict['driver_version'] = VERSION

    def get_network_allocations_number(self):
        """Returns number of network allocations for creating VIFs."""
        # TODO(Shaun Edwards)
        return 0

    def setup_server(self, network_info, metadata=None):
        """Set up and configures share server with given network parameters."""
        # TODO(Shaun Edwards): Look into supporting share servers

    def teardown_server(self, server_details, security_services=None):
        """Teardown share server."""
        # TODO(Shaun Edwards): Look into supporting share servers

    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        """Update share access."""
        if share['share_proto'] == 'NFS':
            state_map = self._update_access_nfs(share, access_rules)
        if share['share_proto'] == 'CIFS':
            state_map = self._update_access_cifs(share, access_rules)
        return state_map

    def _update_access_nfs(self, share, access_rules):
        """Updates access on a NFS share."""
        nfs_rw_ips = set()
        nfs_ro_ips = set()
        rule_state_map = {}
        for rule in access_rules:
            rule_state_map[rule['access_id']] = {
                'state': 'error'
            }

        for rule in access_rules:
            if rule['access_level'] == const.ACCESS_LEVEL_RW:
                nfs_rw_ips.add(rule['access_to'])
            elif rule['access_level'] == const.ACCESS_LEVEL_RO:
                nfs_ro_ips.add(rule['access_to'])

        export_id = self._isilon_api.lookup_nfs_export(
            self._get_container_path(share))
        if export_id is None:
            # share does not exist on backend (set all rules to error state)
            return rule_state_map
        data = {
            'clients': list(nfs_rw_ips),
            'read_only_clients': list(nfs_ro_ips)
        }
        url = ('{0}/platform/1/protocols/nfs/exports/{1}'
               .format(self._server_url, six.text_type(export_id)))
        r = self._isilon_api.request('PUT', url, data=data)
        try:
            r.raise_for_status()
        except HTTPError:
            return rule_state_map

        # if we finish the bulk rule update with no error set rules to active
        for rule in access_rules:
            rule_state_map[rule['access_id']]['state'] = 'active'
        return rule_state_map

    def _update_access_cifs(self, share, access_rules):
        """Clear access on a CIFS share."""
        cifs_ip_set = set()
        users = set()
        for rule in access_rules:
            if rule['access_type'] == 'ip':
                cifs_ip_set.add('allow:' + rule['access_to'])
            elif rule['access_type'] == 'user':
                users.add(rule['access_to'])

        smb_share = self._isilon_api.lookup_smb_share(share['name'])

        backend_smb_user_permissions = smb_share['permissions']
        perms_to_remove = []
        for perm in backend_smb_user_permissions:
            if perm['trustee']['name'] not in users:
                perms_to_remove.append(perm)
        for perm in perms_to_remove:
            backend_smb_user_permissions.remove(perm)

        data = {
            'host_acl': list(cifs_ip_set),
            'permissions': backend_smb_user_permissions,
        }

        url = ('{0}/platform/1/protocols/smb/shares/{1}'
               .format(self._server_url, share['name']))
        r = self._isilon_api.request('PUT', url, data=data)
        try:
            r.raise_for_status()
        except HTTPError:
            # clear access rules failed so set all access rules to error state
            rule_state_map = {}
            for rule in access_rules:
                rule_state_map[rule['access_id']] = {
                    'state': 'error'
                }
            return rule_state_map

        # add access rules that don't exist on backend
        rule_state_map = {}
        for rule in access_rules:
            rule_state_map[rule['access_id']] = {
                'state': 'error'
            }
            try:
                if rule['access_type'] == 'ip':
                    self._cifs_allow_access_ip(rule['access_to'], share,
                                               rule['access_level'])
                    rule_state_map[rule['access_id']]['state'] = 'active'
                elif rule['access_type'] == 'user':
                    backend_users = set()
                    for perm in backend_smb_user_permissions:
                        backend_users.add(perm['trustee']['name'])
                    if rule['access_to'] not in backend_users:
                        self._cifs_allow_access_user(
                            rule['access_to'], share, rule['access_level'])
                    rule_state_map[rule['access_id']]['state'] = 'active'
                else:
                    continue
            except exception.ManilaException:
                pass
        return rule_state_map



# Copyright 2010 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ipaddress
import os
import re
import six
import string

from operator import xor
from oslo_config import cfg
from oslo_log import log
from oslo_utils import encodeutils
from oslo_utils import strutils
from six.moves.urllib import parse
import webob

from manila.api.openstack import api_version_request as api_version
from manila.api.openstack import versioned_method
from manila.common import constants
from manila import exception
from manila.i18n import _
from manila import policy

api_common_opts = [
    cfg.IntOpt(
        'osapi_max_limit',
        default=1000,
        help='The maximum number of items returned in a single response from '
             'a collection resource.'),
    cfg.StrOpt(
        'osapi_share_base_URL',
        help='Base URL to be presented to users in links to the Share API'),
]

CONF = cfg.CONF
CONF.register_opts(api_common_opts)
LOG = log.getLogger(__name__)


# Regex that matches alphanumeric characters, periods, hypens,
# colons and underscores:
# ^ assert position at start of the string
# [\w\.\-\:\_] match expression
# $ assert position at end of the string
VALID_KEY_NAME_REGEX = re.compile(r"^[\w\.\-\:\_]+$", re.UNICODE)


def validate_key_names(key_names_list):
    """Validate each item of the list to match key name regex."""
    for key_name in key_names_list:
        if not VALID_KEY_NAME_REGEX.match(key_name):
            return False
    return True


def get_pagination_params(request):
    """Return marker, limit, offset tuple from request.

    :param request: `wsgi.Request` possibly containing 'marker' and 'limit'
                    GET variables. 'marker' is the id of the last element
                    the client has seen, and 'limit' is the maximum number
                    of items to return. If 'limit' is not specified, 0, or
                    > max_limit, we default to max_limit. Negative values
                    for either marker or limit will cause
                    exc.HTTPBadRequest() exceptions to be raised.

    """
    params = {}
    if 'limit' in request.GET:
        params['limit'] = _get_limit_param(request)
    if 'marker' in request.GET:
        params['marker'] = _get_marker_param(request)
    if 'offset' in request.GET:
        params['offset'] = _get_offset_param(request)
    return params


def _get_limit_param(request):
    """Extract integer limit from request or fail.

    Defaults to max_limit if not present and returns max_limit if present
   'limit' is greater than max_limit.
    """
    max_limit = CONF.osapi_max_limit
    try:
        limit = int(request.GET['limit'])
    except ValueError:
        msg = _('limit param must be an integer')
        raise webob.exc.HTTPBadRequest(explanation=msg)
    if limit < 0:
        msg = _('limit param must be positive')
        raise webob.exc.HTTPBadRequest(explanation=msg)
    limit = min(limit, max_limit)
    return limit


def _get_marker_param(request):
    """Extract marker ID from request or fail."""
    return request.GET['marker']


def _get_offset_param(request):
    """Extract offset id from request's dictionary (defaults to 0) or fail."""
    offset = request.GET['offset']
    return _validate_integer(offset,
                             'offset',
                             0,
                             constants.DB_MAX_INT)


def _validate_integer(value, name, min_value=None, max_value=None):
    """Make sure that value is a valid integer, potentially within range.

    :param value: the value of the integer
    :param name: the name of the integer
    :param min_value: the min_length of the integer
    :param max_value: the max_length of the integer
    :return: integer
    """
    try:
        value = strutils.validate_integer(value, name, min_value, max_value)
        return value
    except ValueError as e:
        raise webob.exc.HTTPBadRequest(explanation=e)


def _validate_pagination_query(request, max_limit=CONF.osapi_max_limit):
    """Validate the given request query and return limit and offset."""

    try:
        offset = int(request.GET.get('offset', 0))
    except ValueError:
        msg = _('offset param must be an integer')
        raise webob.exc.HTTPBadRequest(explanation=msg)

    try:
        limit = int(request.GET.get('limit', max_limit))
    except ValueError:
        msg = _('limit param must be an integer')
        raise webob.exc.HTTPBadRequest(explanation=msg)

    if limit < 0:
        msg = _('limit param must be positive')
        raise webob.exc.HTTPBadRequest(explanation=msg)

    if offset < 0:
        msg = _('offset param must be positive')
        raise webob.exc.HTTPBadRequest(explanation=msg)

    return limit, offset


def limited(items, request, max_limit=CONF.osapi_max_limit):
    """Return a slice of items according to requested offset and limit.

    :param items: A sliceable entity
    :param request: ``wsgi.Request`` possibly containing 'offset' and 'limit'
                    GET variables. 'offset' is where to start in the list,
                    and 'limit' is the maximum number of items to return. If
                    'limit' is not specified, 0, or > max_limit, we default
                    to max_limit. Negative values for either offset or limit
                    will cause exc.HTTPBadRequest() exceptions to be raised.
    :kwarg max_limit: The maximum number of items to return from 'items'
    """
    limit, offset = _validate_pagination_query(request, max_limit)

    limit = min(max_limit, limit or max_limit)
    range_end = offset + limit
    return items[offset:range_end]


def get_sort_params(params, default_key='created_at', default_dir='desc'):
    """Retrieves sort key/direction parameters.

    Processes the parameters to get the 'sort_key' and 'sort_dir' parameter
    values.

    :param params: webob.multidict of request parameters (from
                   manila.api.openstack.wsgi.Request.params)
    :param default_key: default sort key value, will return if no
                        sort key are supplied
    :param default_dir: default sort dir value, will return if no
                        sort dir are supplied
    :returns: value of sort key, value of sort dir
    """
    sort_key = params.pop('sort_key', default_key)
    sort_dir = params.pop('sort_dir', default_dir)
    return sort_key, sort_dir


def remove_version_from_href(href):
    """Removes the first api version from the href.

    Given: 'http://manila.example.com/v1.1/123'
    Returns: 'http://manila.example.com/123'

    Given: 'http://www.manila.com/v1.1'
    Returns: 'http://www.manila.com'

    Given: 'http://manila.example.com/share/v1.1/123'
    Returns: 'http://manila.example.com/share/123'

    """
    parsed_url = parse.urlsplit(href)
    url_parts = parsed_url.path.split('/')

    # NOTE: this should match vX.X or vX
    expression = re.compile(r'^v([0-9]+|[0-9]+\.[0-9]+)(/.*|$)')
    for x in range(len(url_parts)):
        if expression.match(url_parts[x]):
            del url_parts[x]
            break

    new_path = '/'.join(url_parts)

    if new_path == parsed_url.path:
        msg = 'href %s does not contain version' % href
        LOG.debug(msg)
        raise ValueError(msg)

    parsed_url = list(parsed_url)
    parsed_url[2] = new_path
    return parse.urlunsplit(parsed_url)


def dict_to_query_str(params):
    # TODO(throughnothing): we should just use urllib.urlencode instead of this
    # But currently we don't work with urlencoded url's
    param_str = ""
    for key, val in params.items():
        param_str = param_str + '='.join([str(key), str(val)]) + '&'

    return param_str.rstrip('&')


def check_net_id_and_subnet_id(body):
    if xor('neutron_net_id' in body, 'neutron_subnet_id' in body):
        msg = _("When creating a new share network subnet you need to "
                "specify both neutron_net_id and neutron_subnet_id or "
                "none of them.")
        raise webob.exc.HTTPBadRequest(explanation=msg)


def check_share_network_is_active(share_network):
    network_status = share_network.get('status')
    if network_status != constants.STATUS_NETWORK_ACTIVE:
        msg = _("The share network %(id)s used isn't in an 'active' state. "
                "Current status is %(status)s. The action may be retried "
                "after the share network has changed its state.") % {
            'id': share_network['id'],
            'status': share_network.get('status'),
        }
        raise webob.exc.HTTPBadRequest(explanation=msg)


class ViewBuilder(object):
    """Model API responses as dictionaries."""

    _collection_name = None
    _detail_version_modifiers = []

    def _get_project_id(self, request):
        project_id = request.environ["manila.context"].project_id
        if '/v1/' in request.url:
            # project_ids are mandatory in v1 URLs
            return project_id
        elif project_id and ("/v2/%s" % project_id in request.url):
            # project_ids are not mandatory within v2 URLs, but links need
            # to include them if the request does.
            return project_id
        return ''

    def _get_links(self, request, identifier):
        return [{"rel": "self",
                 "href": self._get_href_link(request, identifier), },
                {"rel": "bookmark",
                 "href": self._get_bookmark_link(request, identifier), }]

    def _get_next_link(self, request, identifier):
        """Return href string with proper limit and marker params."""
        params = request.params.copy()
        params["marker"] = identifier
        prefix = self._update_link_prefix(request.application_url,
                                          CONF.osapi_share_base_URL)
        url = os.path.join(prefix,
                           self._get_project_id(request),
                           self._collection_name)
        return "%s?%s" % (url, dict_to_query_str(params))

    def _get_href_link(self, request, identifier):
        """Return an href string pointing to this object."""
        prefix = self._update_link_prefix(request.application_url,
                                          CONF.osapi_share_base_URL)
        return os.path.join(prefix,
                            self._get_project_id(request),
                            self._collection_name,
                            str(identifier))

    def _get_bookmark_link(self, request, identifier):
        """Create a URL that refers to a specific resource."""
        base_url = remove_version_from_href(request.application_url)
        base_url = self._update_link_prefix(base_url,
                                            CONF.osapi_share_base_URL)
        return os.path.join(base_url,
                            self._get_project_id(request),
                            self._collection_name,
                            str(identifier))

    def _get_collection_links(self, request, items, id_key="uuid"):
        """Retrieve 'next' link, if applicable."""
        links = []
        limit = int(request.params.get("limit", 0))
        if limit and limit == len(items):
            last_item = items[-1]
            if id_key in last_item:
                last_item_id = last_item[id_key]
            else:
                last_item_id = last_item["id"]
            links.append({
                "rel": "next",
                "href": self._get_next_link(request, last_item_id),
            })
        return links

    def _update_link_prefix(self, orig_url, prefix):
        if not prefix:
            return orig_url
        url_parts = list(parse.urlsplit(orig_url))
        prefix_parts = list(parse.urlsplit(prefix))
        url_parts[0:2] = prefix_parts[0:2]
        return parse.urlunsplit(url_parts)

    def update_versioned_resource_dict(self, request, resource_dict, resource):
        """Updates the given resource dict for the given request version.

        This method calls every method, that is applicable to the request
        version, in _detail_version_modifiers.
        """
        for method_name in self._detail_version_modifiers:
            method = getattr(self, method_name)
            if request.api_version_request.matches_versioned_method(method):
                request_context = request.environ['manila.context']
                method.func(self, request_context, resource_dict, resource)

    @classmethod
    def versioned_method(cls, min_ver, max_ver=None, experimental=False):
        """Decorator for versioning API methods.

        :param min_ver: string representing minimum version
        :param max_ver: optional string representing maximum version
        :param experimental: flag indicating an API is experimental and is
                             subject to change or removal at any time
        """

        def decorator(f):
            obj_min_ver = api_version.APIVersionRequest(min_ver)
            if max_ver:
                obj_max_ver = api_version.APIVersionRequest(max_ver)
            else:
                obj_max_ver = api_version.APIVersionRequest()

            # Add to list of versioned methods registered
            func_name = f.__name__
            new_func = versioned_method.VersionedMethod(
                func_name, obj_min_ver, obj_max_ver, experimental, f)

            return new_func

        return decorator


def remove_invalid_options(context, search_options, allowed_search_options):
    """Remove search options that are not valid for non-admin API/context."""
    if context.is_admin:
        # Allow all options
        return
    # Otherwise, strip out all unknown options
    unknown_options = [opt for opt in search_options
                       if opt not in allowed_search_options]
    bad_options = ", ".join(unknown_options)
    LOG.debug("Removing options '%(bad_options)s' from query",
              {"bad_options": bad_options})
    for opt in unknown_options:
        del search_options[opt]


def validate_common_name(access):
    """Validate common name passed by user.

    'access' is used as the certificate's CN (common name)
    to which access is allowed or denied by the backend.
    The standard allows for just about any string in the
    common name. The meaning of a string depends on its
    interpretation and is limited to 64 characters.
    """
    if not(0 < len(access) < 65):
        exc_str = _('Invalid CN (common name). Must be 1-64 chars long.')
        raise webob.exc.HTTPBadRequest(explanation=exc_str)


'''
for the reference specification for AD usernames, reference below links:

  1:https://msdn.microsoft.com/en-us/library/bb726984.aspx
  2:https://technet.microsoft.com/en-us/library/cc733146.aspx
'''


def validate_username(access):
    sole_periods_spaces_re = r'[\s|\.]+$'
    valid_username_re = r'.[^\"\/\\\[\]\:\;\|\=\,\+\*\?\<\>]{3,254}$'
    username = access

    if re.match(sole_periods_spaces_re, username):
        exc_str = ('Invalid user or group name,cannot consist solely '
                   'of periods or spaces.')
        raise webob.exc.HTTPBadRequest(explanation=exc_str)

    if not re.match(valid_username_re, username):
        exc_str = ('Invalid user or group name. Must be 4-255 characters '
                   'and consist of alphanumeric characters and '
                   'exclude special characters "/\\[]:;|=,+*?<>')
        raise webob.exc.HTTPBadRequest(explanation=exc_str)


def validate_cephx_id(cephx_id):
    if not cephx_id:
        raise webob.exc.HTTPBadRequest(explanation=_(
            'Ceph IDs may not be empty.'))

    # This restriction may be lifted in Ceph in the future:
    # http://tracker.ceph.com/issues/14626
    if not set(cephx_id) <= set(string.printable):
        raise webob.exc.HTTPBadRequest(explanation=_(
            'Ceph IDs must consist of ASCII printable characters.'))

    # Periods are technically permitted, but we restrict them here
    # to avoid confusion where users are unsure whether they should
    # include the "client." prefix: otherwise they could accidentally
    # create "client.client.foobar".
    if '.' in cephx_id:
        raise webob.exc.HTTPBadRequest(explanation=_(
            'Ceph IDs may not contain periods.'))


def validate_ip(access_to, enable_ipv6):
    try:
        if enable_ipv6:
            validator = ipaddress.ip_network
        else:
            validator = ipaddress.IPv4Network
        validator(six.text_type(access_to))
    except ValueError as error:
        err_msg = encodeutils.exception_to_unicode(error)
        raise webob.exc.HTTPBadRequest(explanation=err_msg)


def validate_access(*args, **kwargs):

    access_type = kwargs.get('access_type')
    access_to = kwargs.get('access_to')
    enable_ceph = kwargs.get('enable_ceph')
    enable_ipv6 = kwargs.get('enable_ipv6')

    if access_type == 'ip':
        validate_ip(access_to, enable_ipv6)
    elif access_type == 'user':
        validate_username(access_to)
    elif access_type == 'cert':
        validate_common_name(access_to.strip())
    elif access_type == "cephx" and enable_ceph:
        validate_cephx_id(access_to)
    else:
        if enable_ceph:
            exc_str = _("Only 'ip', 'user', 'cert' or 'cephx' access "
                        "types are supported.")
        else:
            exc_str = _("Only 'ip', 'user' or 'cert' access types "
                        "are supported.")

        raise webob.exc.HTTPBadRequest(explanation=exc_str)


def validate_integer(value, name, min_value=None, max_value=None):
    """Make sure that value is a valid integer, potentially within range.

    :param value: the value of the integer
    :param name: the name of the integer
    :param min_length: the min_length of the integer
    :param max_length: the max_length of the integer
    :returns: integer
    """
    try:
        value = strutils.validate_integer(value, name, min_value, max_value)
        return value
    except ValueError as e:
        raise webob.exc.HTTPBadRequest(explanation=str(e))


def validate_public_share_policy(context, api_params, api='create'):
    """Validates if policy allows is_public parameter to be set to True.

    :arg api_params - A dictionary of values that may contain 'is_public'
    :returns api_params with 'is_public' item sanitized if present
    :raises exception.InvalidParameterValue if is_public is set but is Invalid
            exception.NotAuthorized if is_public is True but policy prevents it
    """
    if 'is_public' not in api_params:
        return api_params

    policies = {
        'create': 'create_public_share',
        'update': 'set_public_share',
    }
    policy_to_check = policies[api]
    try:
        api_params['is_public'] = strutils.bool_from_string(
            api_params['is_public'], strict=True)
    except ValueError as e:
        raise exception.InvalidParameterValue(six.text_type(e))

    public_shares_allowed = policy.check_policy(
        context, 'share', policy_to_check, do_raise=False)
    if api_params['is_public'] and not public_shares_allowed:
        message = _("User is not authorized to set 'is_public' to True in the "
                    "request.")
        raise exception.NotAuthorized(message=message)

    return api_params



# Copyright 2015 Alex Meade
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manila.api import common


class ShareGroupViewBuilder(common.ViewBuilder):
    """Model a share group API response as a python dictionary."""

    _collection_name = 'share_groups'
    _detail_version_modifiers = [
        "add_consistent_snapshot_support_and_az_id_fields_to_sg",
    ]

    def summary_list(self, request, share_groups):
        """Show a list of share groups without many details."""
        return self._list_view(self.summary, request, share_groups)

    def detail_list(self, request, share_groups):
        """Detailed view of a list of share groups."""
        return self._list_view(self.detail, request, share_groups)

    def summary(self, request, share_group):
        """Generic, non-detailed view of a share group."""
        return {
            'share_group': {
                'id': share_group.get('id'),
                'name': share_group.get('name'),
                'links': self._get_links(request, share_group['id'])
            }
        }

    def detail(self, request, share_group):
        """Detailed view of a single share group."""
        context = request.environ['manila.context']
        share_group_dict = {
            'id': share_group.get('id'),
            'name': share_group.get('name'),
            'created_at': share_group.get('created_at'),
            'status': share_group.get('status'),
            'description': share_group.get('description'),
            'project_id': share_group.get('project_id'),
            'host': share_group.get('host'),
            'share_group_type_id': share_group.get('share_group_type_id'),
            'source_share_group_snapshot_id': share_group.get(
                'source_share_group_snapshot_id'),
            'share_network_id': share_group.get('share_network_id'),
            'share_types': [st['share_type_id'] for st in share_group.get(
                'share_types')],
            'links': self._get_links(request, share_group['id']),
        }
        self.update_versioned_resource_dict(
            request, share_group_dict, share_group)
        if context.is_admin:
            share_group_dict['share_server_id'] = share_group.get(
                'share_server_id')
        return {'share_group': share_group_dict}

    @common.ViewBuilder.versioned_method("2.34")
    def add_consistent_snapshot_support_and_az_id_fields_to_sg(
            self, context, sg_dict, sg):
        sg_dict['availability_zone'] = sg.get('availability_zone')
        sg_dict['consistent_snapshot_support'] = sg.get(
            'consistent_snapshot_support')

    def _list_view(self, func, request, shares):
        """Provide a view for a list of share groups."""
        share_group_list = [
            func(request, share)['share_group']
            for share in shares
        ]
        share_groups_links = self._get_collection_links(
            request, shares, self._collection_name)
        share_groups_dict = {"share_groups": share_group_list}

        if share_groups_links:
            share_groups_dict['share_groups_links'] = share_groups_links

        return share_groups_dict



# Copyright (c) 2016 Hitachi Data Systems
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt

from manila.api.v2 import share_snapshot_export_locations as export_locations
from manila.common import constants
from manila import context
from manila.db.sqlalchemy import api as db_api
from manila import exception
from manila import test
from manila.tests.api import fakes
from manila.tests import db_utils


@ddt.ddt
class ShareSnapshotExportLocationsAPITest(test.TestCase):

    def _get_request(self, version="2.32", use_admin_context=True):
        req = fakes.HTTPRequest.blank(
            '/v2/fake/snapshots/%s/export-locations' % self.snapshot['id'],
            version=version, use_admin_context=use_admin_context)
        return req

    def setUp(self):
        super(ShareSnapshotExportLocationsAPITest, self).setUp()
        self.controller = (
            export_locations.ShareSnapshotExportLocationController())

        self.share = db_utils.create_share()
        self.snapshot = db_utils.create_snapshot(
            status=constants.STATUS_AVAILABLE,
            share_id=self.share['id'])
        self.snapshot_instance = db_utils.create_snapshot_instance(
            status=constants.STATUS_AVAILABLE,
            share_instance_id=self.share['instance']['id'],
            snapshot_id=self.snapshot['id'])

        self.values = {
            'share_snapshot_instance_id': self.snapshot_instance['id'],
            'path': 'fake/user_path',
            'is_admin_only': True,
        }

        self.exp_loc = db_api.share_snapshot_instance_export_location_create(
            context.get_admin_context(), self.values)

        self.req = self._get_request()

    def test_index(self):
        self.mock_object(
            db_api, 'share_snapshot_instance_export_locations_get_all',
            mock.Mock(return_value=[self.exp_loc]))
        out = self.controller.index(self._get_request('2.32'),
                                    self.snapshot['id'])

        values = {
            'share_snapshot_export_locations': [{
                'share_snapshot_instance_id': self.snapshot_instance['id'],
                'path': 'fake/user_path',
                'is_admin_only': True,
                'id': self.exp_loc['id'],
                'links': [{
                    'href': 'http://localhost/share/v2/fake/'
                            'share_snapshot_export_locations/' +
                            self.exp_loc['id'],
                    'rel': 'self'
                }, {
                    'href': 'http://localhost/share/fake/'
                            'share_snapshot_export_locations/' +
                            self.exp_loc['id'],
                    'rel': 'bookmark'
                }],

            }]
        }
        self.assertSubDictMatch(values, out)

    def test_show(self):
        out = self.controller.show(self._get_request('2.32'),
                                   self.snapshot['id'], self.exp_loc['id'])

        self.assertSubDictMatch(
            {'share_snapshot_export_location': self.values}, out)

    @ddt.data('1.0', '2.0', '2.5', '2.8', '2.31')
    def test_list_with_unsupported_version(self, version):
        self.assertRaises(
            exception.VersionNotFoundForAPIMethod,
            self.controller.index,
            self._get_request(version),
            self.snapshot_instance['id'],
        )

    @ddt.data('1.0', '2.0', '2.5', '2.8', '2.31')
    def test_show_with_unsupported_version(self, version):
        self.assertRaises(
            exception.VersionNotFoundForAPIMethod,
            self.controller.show,
            self._get_request(version),
            self.snapshot['id'],
            self.exp_loc['id']
        )



# Copyright 2019 Nexenta by DDN, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import posixpath

from oslo_log import log
from oslo_utils import units

from manila.common import constants as common
from manila import exception
from manila.i18n import _
from manila.share import driver
from manila.share.drivers.nexenta.ns5 import jsonrpc
from manila.share.drivers.nexenta import options
from manila.share.drivers.nexenta import utils

VERSION = '1.1'
LOG = log.getLogger(__name__)
ZFS_MULTIPLIER = 1.1  # ZFS quotas do not take metadata into account.


class NexentaNasDriver(driver.ShareDriver):
    """Nexenta Share Driver.

    Executes commands relating to Shares.
    API version history:
        1.0 - Initial version.
        1.1 - Failover support.
            - Unshare filesystem completely after last securityContext
              is removed.
            - Moved all http/url code to jsonrpc.
            - Manage existing support.
            - Revert to snapshot support.
    """

    driver_prefix = 'nexenta'

    def __init__(self, *args, **kwargs):
        """Do initialization."""
        LOG.debug('Initializing Nexenta driver.')
        super(NexentaNasDriver, self).__init__(False, *args, **kwargs)
        self.configuration = kwargs.get('configuration')
        if self.configuration:
            self.configuration.append_config_values(
                options.nexenta_connection_opts)
            self.configuration.append_config_values(
                options.nexenta_nfs_opts)
            self.configuration.append_config_values(
                options.nexenta_dataset_opts)
        else:
            raise exception.BadConfigurationException(
                reason=_('Nexenta configuration missing.'))

        self.nef = None
        self.verify_ssl = self.configuration.nexenta_ssl_cert_verify
        self.nas_host = self.configuration.nexenta_nas_host
        self.nef_port = self.configuration.nexenta_rest_port
        self.nef_user = self.configuration.nexenta_user
        self.nef_password = self.configuration.nexenta_password

        self.pool_name = self.configuration.nexenta_pool
        self.parent_fs = self.configuration.nexenta_folder

        self.nfs_mount_point_base = self.configuration.nexenta_mount_point_base
        self.dataset_compression = (
            self.configuration.nexenta_dataset_compression)
        self.provisioned_capacity = 0

    @property
    def storage_protocol(self):
        protocol = ''
        if self.configuration.nexenta_nfs:
            protocol = 'NFS'
        else:
            msg = _('At least 1 storage protocol must be enabled.')
            raise exception.NexentaException(msg)
        return protocol

    @property
    def root_path(self):
        return posixpath.join(self.pool_name, self.parent_fs)

    @property
    def share_backend_name(self):
        if not hasattr(self, '_share_backend_name'):
            self._share_backend_name = None
            if self.configuration:
                self._share_backend_name = self.configuration.safe_get(
                    'share_backend_name')
            if not self._share_backend_name:
                self._share_backend_name = 'NexentaStor5'
        return self._share_backend_name

    def do_setup(self, context):
        self.nef = jsonrpc.NefProxy(self.storage_protocol,
                                    self.root_path,
                                    self.configuration)

    def check_for_setup_error(self):
        """Check root filesystem, NFS service and NFS share."""
        filesystem = self.nef.filesystems.get(self.root_path)
        if filesystem['mountPoint'] == 'none':
            message = (_('NFS root filesystem %(path)s is not writable')
                       % {'path': filesystem['mountPoint']})
            raise jsonrpc.NefException(code='ENOENT', message=message)
        if not filesystem['isMounted']:
            message = (_('NFS root filesystem %(path)s is not mounted')
                       % {'path': filesystem['mountPoint']})
            raise jsonrpc.NefException(code='ENOTDIR', message=message)
        payload = {}
        if filesystem['nonBlockingMandatoryMode']:
            payload['nonBlockingMandatoryMode'] = False
        if filesystem['smartCompression']:
            payload['smartCompression'] = False
        if payload:
            self.nef.filesystems.set(self.root_path, payload)
        service = self.nef.services.get('nfs')
        if service['state'] != 'online':
            message = (_('NFS server service is not online: %(state)s')
                       % {'state': service['state']})
            raise jsonrpc.NefException(code='ESRCH', message=message)
        self._get_provisioned_capacity()

    def _get_provisioned_capacity(self):
        payload = {'fields': 'referencedQuotaSize'}
        self.provisioned_capacity += self.nef.filesystems.get(
            self.root_path, payload)['referencedQuotaSize']

    def ensure_share(self, context, share, share_server=None):
        pass

    def create_share(self, context, share, share_server=None):
        """Create a share."""
        LOG.debug('Creating share: %s.', self._get_share_name(share))
        dataset_path = self._get_dataset_path(share)
        size = int(share['size'] * units.Gi * ZFS_MULTIPLIER)
        payload = {
            'recordSize': self.configuration.nexenta_dataset_record_size,
            'compressionMode': self.dataset_compression,
            'path': dataset_path,
            'referencedQuotaSize': size,
            'nonBlockingMandatoryMode': False
        }
        if not self.configuration.nexenta_thin_provisioning:
            payload['referencedReservationSize'] = size
        self.nef.filesystems.create(payload)

        try:
            mount_path = self._mount_filesystem(share)
        except jsonrpc.NefException as create_error:
            try:
                payload = {'force': True}
                self.nef.filesystems.delete(dataset_path, payload)
            except jsonrpc.NefException as delete_error:
                LOG.debug('Failed to delete share %(path)s: %(error)s',
                          {'path': dataset_path, 'error': delete_error})
            raise create_error

        self.provisioned_capacity += share['size']
        location = {
            'path': mount_path,
            'id': self._get_share_name(share)
        }
        return [location]

    def _mount_filesystem(self, share):
        """Ensure that filesystem is activated and mounted on the host."""
        dataset_path = self._get_dataset_path(share)
        payload = {'fields': 'mountPoint,isMounted'}
        filesystem = self.nef.filesystems.get(dataset_path, payload)
        if filesystem['mountPoint'] == 'none':
            payload = {'datasetName': dataset_path}
            self.nef.hpr.activate(payload)
            filesystem = self.nef.filesystems.get(dataset_path, payload)
        elif not filesystem['isMounted']:
            self.nef.filesystems.mount(dataset_path)
        return '%s:%s' % (self.nas_host, filesystem['mountPoint'])

    def create_share_from_snapshot(self, context, share, snapshot,
                                   share_server=None, parent_share=None):
        """Is called to create share from snapshot."""
        snapshot_path = self._get_snapshot_path(snapshot)
        LOG.debug('Creating share from snapshot %s.', snapshot_path)
        clone_path = self._get_dataset_path(share)
        size = int(share['size'] * units.Gi * ZFS_MULTIPLIER)
        payload = {
            'targetPath': clone_path,
            'referencedQuotaSize': size,
            'recordSize': self.configuration.nexenta_dataset_record_size,
            'compressionMode': self.dataset_compression,
            'nonBlockingMandatoryMode': False
        }
        if not self.configuration.nexenta_thin_provisioning:
            payload['referencedReservationSize'] = size
        self.nef.snapshots.clone(snapshot_path, payload)
        self._remount_filesystem(clone_path)
        self.provisioned_capacity += share['size']
        try:
            mount_path = self._mount_filesystem(share)
        except jsonrpc.NefException as create_error:
            try:
                payload = {'force': True}
                self.nef.filesystems.delete(clone_path, payload)
            except jsonrpc.NefException as delete_error:
                LOG.debug('Failed to delete share %(path)s: %(error)s',
                          {'path': clone_path, 'error': delete_error})
            raise create_error

        location = {
            'path': mount_path,
            'id': self._get_share_name(share)
        }
        return [location]

    def _remount_filesystem(self, clone_path):
        """Workaround for NEF bug: cloned share has offline NFS status"""
        self.nef.filesystems.unmount(clone_path)
        self.nef.filesystems.mount(clone_path)

    def _get_dataset_path(self, share):
        share_name = self._get_share_name(share)
        return posixpath.join(self.root_path, share_name)

    def _get_share_name(self, share):
        """Get share name with share name prefix."""
        return ('%(prefix)s%(share_id)s' % {
                'prefix': self.configuration.nexenta_share_name_prefix,
                'share_id': share['share_id']})

    def _get_snapshot_path(self, snapshot):
        """Return ZFS snapshot path for the snapshot."""
        snapshot_id = (
            snapshot['snapshot_id'] or snapshot['share_group_snapshot_id'])
        share = snapshot.get('share') or snapshot.get('share_instance')
        fs_path = self._get_dataset_path(share)
        return '%s@snapshot-%s' % (fs_path, snapshot_id)

    def delete_share(self, context, share, share_server=None):
        """Delete a share."""
        LOG.debug('Deleting share: %s.', self._get_share_name(share))
        share_path = self._get_dataset_path(share)
        delete_payload = {'force': True, 'snapshots': True}
        try:
            self.nef.filesystems.delete(share_path, delete_payload)
        except jsonrpc.NefException as error:
            if error.code != 'EEXIST':
                raise error
            snapshots_tree = {}
            snapshots_payload = {'parent': share_path, 'fields': 'path'}
            snapshots = self.nef.snapshots.list(snapshots_payload)
            for snapshot in snapshots:
                clones_payload = {'fields': 'clones,creationTxg'}
                data = self.nef.snapshots.get(snapshot['path'], clones_payload)
                if data['clones']:
                    snapshots_tree[data['creationTxg']] = data['clones'][0]
            if snapshots_tree:
                clone_path = snapshots_tree[max(snapshots_tree)]
                self.nef.filesystems.promote(clone_path)
            self.nef.filesystems.delete(share_path, delete_payload)
        self.provisioned_capacity -= share['size']

    def extend_share(self, share, new_size, share_server=None):
        """Extends a share."""
        LOG.debug(
            'Extending share: %(name)s to %(size)sG.', (
                {'name': self._get_share_name(share), 'size': new_size}))
        self._set_quota(share, new_size)
        if not self.configuration.nexenta_thin_provisioning:
            self._set_reservation(share, new_size)
        self.provisioned_capacity += (new_size - share['size'])

    def shrink_share(self, share, new_size, share_server=None):
        """Shrinks size of existing share."""
        LOG.debug(
            'Shrinking share: %(name)s to %(size)sG.', {
                'name': self._get_share_name(share), 'size': new_size})
        share_path = self._get_dataset_path(share)
        share_data = self.nef.filesystems.get(share_path)
        used = share_data['bytesUsedBySelf'] / units.Gi
        if used > new_size:
            raise exception.ShareShrinkingPossibleDataLoss(
                share_id=self._get_share_name(share))
        if not self.configuration.nexenta_thin_provisioning:
            self._set_reservation(share, new_size)
        self._set_quota(share, new_size)
        self.provisioned_capacity += (share['size'] - new_size)

    def create_snapshot(self, context, snapshot, share_server=None):
        """Create a snapshot."""
        snapshot_path = self._get_snapshot_path(snapshot)
        LOG.debug('Creating snapshot: %s.', snapshot_path)
        payload = {'path': snapshot_path}
        self.nef.snapshots.create(payload)

    def delete_snapshot(self, context, snapshot, share_server=None):
        """Deletes a snapshot.

        :param snapshot: snapshot reference
        """
        snapshot_path = self._get_snapshot_path(snapshot)
        LOG.debug('Deleting snapshot: %s.', snapshot_path)
        payload = {'defer': True}
        self.nef.snapshots.delete(snapshot_path, payload)

    def revert_to_snapshot(self, context, snapshot, share_access_rules,
                           snapshot_access_rules, share_server=None):
        """Reverts a share (in place) to the specified snapshot.

        Does not delete the share snapshot.  The share and snapshot must both
        be 'available' for the restore to be attempted.  The snapshot must be
        the most recent one taken by Manila; the API layer performs this check
        so the driver doesn't have to.

        The share must be reverted in place to the contents of the snapshot.
        Application admins should quiesce or otherwise prepare the application
        for the shared file system contents to change suddenly.

        :param context: Current context
        :param snapshot: The snapshot to be restored
        :param share_access_rules: List of all access rules for the affected
            share
        :param snapshot_access_rules: List of all access rules for the affected
            snapshot
        :param share_server: Optional -- Share server model or None
        """
        snapshot_path = self._get_snapshot_path(snapshot).split('@')[1]
        LOG.debug('Reverting to snapshot: %s.', snapshot_path)
        share_path = self._get_dataset_path(snapshot['share'])
        payload = {'snapshot': snapshot_path}
        self.nef.filesystems.rollback(share_path, payload)

    def manage_existing(self, share, driver_options):
        """Brings an existing share under Manila management.

        If the provided share is not valid, then raise a
        ManageInvalidShare exception, specifying a reason for the failure.

        If the provided share is not in a state that can be managed, such as
        being replicated on the backend, the driver *MUST* raise
        ManageInvalidShare exception with an appropriate message.

        The share has a share_type, and the driver can inspect that and
        compare against the properties of the referenced backend share.
        If they are incompatible, raise a
        ManageExistingShareTypeMismatch, specifying a reason for the failure.

        :param share: Share model
        :param driver_options: Driver-specific options provided by admin.
        :return: share_update dictionary with required key 'size',
                 which should contain size of the share.
        """
        LOG.debug('Manage share %s.', self._get_share_name(share))
        export_path = share['export_locations'][0]['path']

        # check that filesystem with provided export exists.
        fs_path = export_path.split(':/')[1]
        fs_data = self.nef.filesystems.get(fs_path)

        if not fs_data:
            # wrong export path, raise exception.
            msg = _('Share %s does not exist on Nexenta Store appliance, '
                    'cannot manage.') % export_path
            raise exception.NexentaException(msg)

        # get dataset properties.
        if fs_data['referencedQuotaSize']:
            size = (fs_data['referencedQuotaSize'] / units.Gi) + 1
        else:
            size = fs_data['bytesReferenced'] / units.Gi + 1
        # rename filesystem on appliance to correlate with manila ID.
        new_path = '%s/%s' % (self.root_path, self._get_share_name(share))
        self.nef.filesystems.rename(fs_path, {'newPath': new_path})
        # make sure quotas and reservations are correct.
        if not self.configuration.nexenta_thin_provisioning:
            self._set_reservation(share, size)
        self._set_quota(share, size)

        return {'size': size, 'export_locations': [{
            'path': '%s:/%s' % (self.nas_host, new_path)
        }]}

    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        """Update access rules for given share.

        Using access_rules list for both adding and deleting rules.
        :param context: The `context.RequestContext` object for the request
        :param share: Share that will have its access rules updated.
        :param access_rules: All access rules for given share. This list
        is enough to update the access rules for given share.
        :param add_rules: Empty List or List of access rules which should be
        added. access_rules already contains these rules. Not used by this
        driver.
        :param delete_rules: Empty List or List of access rules which should be
        removed. access_rules doesn't contain these rules. Not used by
        this driver.
        :param share_server: Data structure with share server information.
        Not used by this driver.
        """
        LOG.debug('Updating access to share %(id)s with following access '
                  'rules: %(rules)s', {
                      'id': self._get_share_name(share),
                      'rules': [(
                          rule.get('access_type'), rule.get('access_level'),
                          rule.get('access_to')) for rule in access_rules]})
        rw_list = []
        ro_list = []
        update_dict = {}
        if share['share_proto'] == 'NFS':
            for rule in access_rules:
                if rule['access_type'].lower() != 'ip':
                    msg = _(
                        'Only IP access control type is supported for NFS.')
                    LOG.warning(msg)
                    update_dict[rule['access_id']] = {
                        'state': 'error',
                    }
                else:
                    update_dict[rule['access_id']] = {
                        'state': 'active',
                    }
                if rule['access_level'] == common.ACCESS_LEVEL_RW:
                    rw_list.append(rule['access_to'])
                else:
                    ro_list.append(rule['access_to'])
            self._update_nfs_access(share, rw_list, ro_list)
        return update_dict

    def _update_nfs_access(self, share, rw_list, ro_list):
        # Define allowed security context types to be able to tell whether
        # the 'security_contexts' dict contains any rules at all
        context_types = {'none', 'root', 'readOnlyList', 'readWriteList'}

        security_contexts = {'securityModes': ['sys']}

        def add_sc(addr_list, sc_type):
            if sc_type not in context_types:
                return

            rule_list = []

            for addr in addr_list:
                address_mask = addr.strip().split('/', 1)
                address = address_mask[0]
                ls = {"allow": True, "etype": "fqdn", "entity": address}
                if len(address_mask) == 2:
                    mask = int(address_mask[1])
                    if 0 <= mask < 31:
                        ls['mask'] = mask
                        ls['etype'] = 'network'
                rule_list.append(ls)

            # Context type with no addresses will result in an API error
            if rule_list:
                security_contexts[sc_type] = rule_list

        add_sc(rw_list, 'readWriteList')
        add_sc(ro_list, 'readOnlyList')
        payload = {'securityContexts': [security_contexts]}
        share_path = self._get_dataset_path(share)
        if self.nef.nfs.list({'filesystem': share_path}):
            if not set(security_contexts.keys()) & context_types:
                self.nef.nfs.delete(share_path)
            else:
                self.nef.nfs.set(share_path, payload)
        else:
            payload['filesystem'] = share_path
            self.nef.nfs.create(payload)
        payload = {
            'flags': ['file_inherit', 'dir_inherit'],
            'permissions': ['full_set'],
            'principal': 'everyone@',
            'type': 'allow'
        }
        self.nef.filesystems.acl(share_path, payload)

    def _set_quota(self, share, new_size):
        quota = int(new_size * units.Gi * ZFS_MULTIPLIER)
        share_path = self._get_dataset_path(share)
        payload = {'referencedQuotaSize': quota}
        LOG.debug('Setting quota for dataset %s.', share_path)
        self.nef.filesystems.set(share_path, payload)

    def _set_reservation(self, share, new_size):
        res_size = int(new_size * units.Gi * ZFS_MULTIPLIER)
        share_path = self._get_dataset_path(share)
        payload = {'referencedReservationSize': res_size}
        self.nef.filesystems.set(share_path, payload)

    def _update_share_stats(self, data=None):
        super(NexentaNasDriver, self)._update_share_stats()
        total, free, allocated = self._get_capacity_info()
        compression = not self.dataset_compression == 'off'
        data = {
            'vendor_name': 'Nexenta',
            'storage_protocol': self.storage_protocol,
            'share_backend_name': self.share_backend_name,
            'nfs_mount_point_base': self.nfs_mount_point_base,
            'driver_version': VERSION,
            'snapshot_support': True,
            'create_share_from_snapshot_support': True,
            'revert_to_snapshot_support': True,
            'pools': [{
                'pool_name': self.pool_name,
                'compression': compression,
                'total_capacity_gb': total,
                'free_capacity_gb': free,
                'reserved_percentage': (
                    self.configuration.reserved_share_percentage),
                'reserved_snapshot_percentage':
                    (self.configuration.reserved_share_from_snapshot_percentage
                        or self.configuration.reserved_share_percentage),
                'max_over_subscription_ratio': (
                    self.configuration.safe_get(
                        'max_over_subscription_ratio')),
                'thin_provisioning':
                    self.configuration.nexenta_thin_provisioning,
                'provisioned_capacity_gb': self.provisioned_capacity,
            }],
        }
        self._stats.update(data)

    def _get_capacity_info(self):
        """Calculate available space on the NFS share."""
        data = self.nef.filesystems.get(self.root_path)
        free = int(utils.bytes_to_gb(data['bytesAvailable']))
        allocated = int(utils.bytes_to_gb(data['bytesUsed']))
        total = free + allocated
        return total, free, allocated



# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Handles all requests to Glance.
"""

from glanceclient import client as glance_client
from keystoneauth1 import loading as ks_loading
from oslo_config import cfg

from manila.common import client_auth
from manila.common.config import core_opts
from manila.db import base

GLANCE_GROUP = 'glance'
AUTH_OBJ = None


glance_opts = [
    cfg.StrOpt('api_microversion',
               default='2',
               help='Version of Glance API to be used.'),
    cfg.StrOpt('region_name',
               default='RegionOne',
               help='Region name for connecting to glance.'),
    ]

CONF = cfg.CONF
CONF.register_opts(core_opts)
CONF.register_opts(glance_opts, GLANCE_GROUP)
ks_loading.register_session_conf_options(CONF, GLANCE_GROUP)
ks_loading.register_auth_conf_options(CONF, GLANCE_GROUP)


def list_opts():
    return client_auth.AuthClientLoader.list_opts(GLANCE_GROUP)


def glanceclient(context):
    global AUTH_OBJ
    if not AUTH_OBJ:
        AUTH_OBJ = client_auth.AuthClientLoader(
            client_class=glance_client.Client, cfg_group=GLANCE_GROUP)
    return AUTH_OBJ.get_client(context,
                               version=CONF[GLANCE_GROUP].api_microversion,
                               region_name=CONF[GLANCE_GROUP].region_name)


class API(base.Base):
    """API for interacting with glanceclient."""

    def image_list(self, context):
        client = glanceclient(context)
        if hasattr(client, 'images'):
            return client.images.list()
        return client.glance.list()



# Copyright 2018 Inspur Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Share driver for Inspur AS13000
"""

import eventlet
import functools
import json
import re
import requests
import six
import time

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import units

from manila import exception
from manila.i18n import _
from manila.share import driver
from manila.share import utils as share_utils


inspur_as13000_opts = [
    cfg.HostAddressOpt(
        'as13000_nas_ip',
        required=True,
        help='IP address for the AS13000 storage.'),
    cfg.PortOpt(
        'as13000_nas_port',
        default=8088,
        help='Port number for the AS13000 storage.'),
    cfg.StrOpt(
        'as13000_nas_login',
        required=True,
        help='Username for the AS13000 storage'),
    cfg.StrOpt(
        'as13000_nas_password',
        required=True,
        secret=True,
        help='Password for the AS13000 storage'),
    cfg.ListOpt(
        'as13000_share_pools',
        required=True,
        help='The Storage Pools Manila should use, a comma separated list'),
    cfg.IntOpt(
        'as13000_token_available_time',
        default=3600,
        help='The effective time of token validity in seconds.')
]

CONF = cfg.CONF
CONF.register_opts(inspur_as13000_opts)
LOG = logging.getLogger(__name__)


def inspur_driver_debug_trace(f):
    """Log the method entrance and exit including active backend name.

    This should only be used on Share_Driver class methods. It depends on
    having a 'self' argument that is a AS13000_Driver.
    """
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        driver = args[0]
        cls_name = driver.__class__.__name__
        method_name = "%(cls_name)s.%(method)s" % {"cls_name": cls_name,
                                                   "method": f.__name__}
        backend_name = driver.configuration.share_backend_name
        LOG.debug("[%(backend_name)s] Enter %(method_name)s",
                  {"method_name": method_name, "backend_name": backend_name})
        result = f(*args, **kwargs)
        LOG.debug("[%(backend_name)s] Leave %(method_name)s",
                  {"method_name": method_name, "backend_name": backend_name})
        return result

    return wrapper


class RestAPIExecutor(object):
    def __init__(self, hostname, port, username, password):
        self._hostname = hostname
        self._port = port
        self._username = username
        self._password = password
        self._token_pool = []
        self._token_size = 1

    def logins(self):
        """login the AS13000 and store the token in token_pool"""
        times = self._token_size
        while times > 0:
            token = self.login()
            self._token_pool.append(token)
            times = times - 1
        LOG.debug('Logged into the AS13000.')

    def login(self):
        """login in the AS13000 and return the token"""
        method = 'security/token'
        params = {'name': self._username, 'password': self._password}
        token = self.send_rest_api(method=method, params=params,
                                   request_type='post').get('token')
        return token

    def logout(self):
        method = 'security/token'
        self.send_rest_api(method=method, request_type='delete')

    def refresh_token(self, force=False):
        if force is True:
            for i in range(self._token_size):
                self._token_pool = []
                token = self.login()
                self._token_pool.append(token)
        else:
            for i in range(self._token_size):
                self.logout()
                token = self.login()
                self._token_pool.append(token)
        LOG.debug('Tokens have been refreshed.')

    def send_rest_api(self, method, params=None, request_type='post'):
        attempts = 3
        msge = ''
        while attempts > 0:
            attempts -= 1
            try:
                return self.send_api(method, params, request_type)
            except exception.NetworkException as e:
                msge = six.text_type(e)
                LOG.error(msge)

                self.refresh_token(force=True)
                eventlet.sleep(1)
            except exception.ShareBackendException as e:
                msge = six.text_type(e)
                break

        msg = (_('Access RestAPI /rest/%(method)s by %(type)s failed,'
                 ' error: %(msge)s') % {'method': method,
                                        'msge': msge,
                                        'type': request_type})
        LOG.error(msg)
        raise exception.ShareBackendException(msg)

    @staticmethod
    def do_request(cmd, url, header, data):
        LOG.debug('CMD: %(cmd)s, URL: %(url)s, DATA: %(data)s',
                  {'cmd': cmd, 'url': url, 'data': data})
        if cmd == 'post':
            req = requests.post(url,
                                data=data,
                                headers=header)
        elif cmd == 'get':
            req = requests.get(url,
                               data=data,
                               headers=header)
        elif cmd == 'put':
            req = requests.put(url,
                               data=data,
                               headers=header)
        elif cmd == 'delete':
            req = requests.delete(url,
                                  data=data,
                                  headers=header)
        else:
            msg = (_('Unsupported cmd: %s') % cmd)
            raise exception.ShareBackendException(msg)

        response = req.json()
        code = req.status_code
        LOG.debug('CODE: %(code)s, RESPONSE: %(response)s',
                  {'code': code, 'response': response})

        if code != 200:
            msg = (_('Code: %(code)s, URL: %(url)s, Message: %(msg)s')
                   % {'code': req.status_code,
                      'url': req.url,
                      'msg': req.text})
            LOG.error(msg)
            raise exception.NetworkException(msg)

        return response

    def send_api(self, method, params=None, request_type='post'):
        if params:
            params = json.dumps(params)

        url = ('http://%(hostname)s:%(port)s/%(rest)s/%(method)s'
               % {'hostname': self._hostname,
                  'port': self._port,
                  'rest': 'rest',
                  'method': method})

        # header is not needed when the driver login the backend
        if method == 'security/token':
            # token won't be return to the token_pool
            if request_type == 'delete':
                header = {'X-Auth-Token': self._token_pool.pop(0)}
            else:
                header = None
        else:
            if len(self._token_pool) == 0:
                self.logins()
            token = self._token_pool.pop(0)
            header = {'X-Auth-Token': token}
            self._token_pool.append(token)

        response = self.do_request(request_type, url, header, params)

        try:
            code = response.get('code')
            if code == 0:
                if request_type == 'get':
                    data = response.get('data')
                else:
                    if method == 'security/token':
                        data = response.get('data')
                    else:
                        data = response.get('message')
                        data = str(data).lower()
                        if hasattr(data, 'success'):
                            return
            elif code == 301:
                msg = _('Token is expired')
                LOG.error(msg)
                raise exception.NetworkException(msg)
            else:
                message = response.get('message')
                msg = (_('Unexpected RestAPI response: %(code)d %(msg)s') % {
                       'code': code, 'msg': message})
                LOG.error(msg)
                raise exception.ShareBackendException(msg)
        except ValueError:
            msg = _("Deal with response failed")
            raise exception.ShareBackendException(msg)

        return data


class AS13000ShareDriver(driver.ShareDriver):

    """AS13000 Share Driver

    Version history:
    V1.0.0:    Initial version
               Driver support:
                   share create/delete,
                   snapshot create/delete,
                   extend size,
                   create_share_from_snapshot,
                   update_access.
                   protocol: NFS/CIFS

    """

    VENDOR = 'INSPUR'
    VERSION = '1.0.0'
    PROTOCOL = 'NFS_CIFS'

    def __init__(self, *args, **kwargs):
        super(AS13000ShareDriver, self).__init__(False, *args, **kwargs)
        self.configuration.append_config_values(inspur_as13000_opts)
        self.hostname = self.configuration.as13000_nas_ip
        self.port = self.configuration.as13000_nas_port
        self.username = self.configuration.as13000_nas_login
        self.password = self.configuration.as13000_nas_password
        self.token_available_time = (self.configuration.
                                     as13000_token_available_time)
        self.pools = self.configuration.as13000_share_pools
        # base dir detail contain the information which we will use
        # when we create subdirectorys
        self.base_dir_detail = None
        self._token_time = 0
        self.ips = []
        self._rest = RestAPIExecutor(self.hostname, self.port,
                                     self.username, self.password)

    @inspur_driver_debug_trace
    def do_setup(self, context):
        # get access tokens
        self._rest.logins()
        self._token_time = time.time()

        # Check the pool in conf exist in the backend
        self._validate_pools_exist()

        # get the base directory detail
        self.base_dir_detail = self._get_directory_detail(self.pools[0])

        # get all backend node ip
        self.ips = self._get_nodes_ips()

    @inspur_driver_debug_trace
    def check_for_setup_error(self):
        if self.base_dir_detail is None:
            msg = _('The pool status is not right')
            raise exception.ShareBackendException(msg)

        if len(self.ips) == 0:
            msg = _('All backend nodes are down')
            raise exception.ShareBackendException(msg)

    @inspur_driver_debug_trace
    def create_share(self, context, share, share_server=None):
        """Create a share."""
        pool, name, size, proto = self._get_share_instance_pnsp(share)

        # create directory first
        share_path = self._create_directory(share_name=name,
                                            pool_name=pool)

        # then create nfs or cifs share
        if proto == 'nfs':
            self._create_nfs_share(share_path=share_path)
        else:
            self._create_cifs_share(share_name=name,
                                    share_path=share_path)

        # finally we set the quota of directory
        self._set_directory_quota(share_path, size)

        locations = self._get_location_path(name, share_path, proto)
        LOG.debug('Create share: name:%(name)s'
                  ' protocol:%(proto)s,location: %(loc)s',
                  {'name': name, 'proto': proto, 'loc': locations})
        return locations

    @inspur_driver_debug_trace
    def create_share_from_snapshot(self, context, share, snapshot,
                                   share_server=None, parent_share=None):
        """Create a share from snapshot."""
        pool, name, size, proto = self._get_share_instance_pnsp(share)

        # create directory first
        share_path = self._create_directory(share_name=name,
                                            pool_name=pool)

        # as quota must be set when directory is empty
        # then we set the quota of directory
        self._set_directory_quota(share_path, size)

        # and next clone snapshot to dest_path
        self._clone_directory_to_dest(snapshot=snapshot, dest_path=share_path)

        # finally create share
        if proto == 'nfs':
            self._create_nfs_share(share_path=share_path)
        else:
            self._create_cifs_share(share_name=name,
                                    share_path=share_path)

        locations = self._get_location_path(name, share_path, proto)
        LOG.debug('Create share from snapshot:'
                  ' name:%(name)s protocol:%(proto)s,location: %(loc)s',
                  {'name': name, 'proto': proto, 'loc': locations})
        return locations

    @inspur_driver_debug_trace
    def delete_share(self, context, share, share_server=None):
        """Delete share."""
        pool, name, _, proto = self._get_share_instance_pnsp(share)
        share_path = self._generate_share_path(pool, name)
        if proto == 'nfs':
            share_backend = self._get_nfs_share(share_path)
            if len(share_backend) == 0:
                return
            else:
                self._delete_nfs_share(share_path)
        else:
            share_backend = self._get_cifs_share(name)
            if len(share_backend) == 0:
                return
            else:
                self._delete_cifs_share(name)
        self._delete_directory(share_path)
        LOG.debug('Delete share: %s', name)

    @inspur_driver_debug_trace
    def extend_share(self, share, new_size, share_server=None):
        """extend share to new size"""
        pool, name, size, proto = self._get_share_instance_pnsp(share)
        share_path = self._generate_share_path(pool, name)
        self._set_directory_quota(share_path, new_size)
        LOG.debug('extend share %(name)s to new size %(size)s GB',
                  {'name': name, 'size': new_size})

    @inspur_driver_debug_trace
    def ensure_share(self, context, share, share_server=None):
        """Ensure that share is exported."""
        pool, name, size, proto = self._get_share_instance_pnsp(share)
        share_path = self._generate_share_path(pool, name)

        if proto == 'nfs':
            share_backend = self._get_nfs_share(share_path)
        elif proto == 'cifs':
            share_backend = self._get_cifs_share(name)
        else:
            msg = (_('Invalid NAS protocol supplied: %s.') % proto)
            LOG.error(msg)
            raise exception.InvalidInput(msg)

        if len(share_backend) == 0:
            raise exception.ShareResourceNotFound(share_id=share['share_id'])

        return self._get_location_path(name, share_path, proto)

    @inspur_driver_debug_trace
    def create_snapshot(self, context, snapshot, share_server=None):
        """create snapshot of share"""
        # !!! Attention the share property is a ShareInstance
        share = snapshot['share']
        pool, share_name, _, _ = self._get_share_instance_pnsp(share)
        share_path = self._generate_share_path(pool, share_name)

        snap_name = self._generate_snapshot_name(snapshot)

        method = 'snapshot/directory'
        request_type = 'post'
        params = {'path': share_path, 'snapName': snap_name}
        self._rest.send_rest_api(method=method,
                                 params=params,
                                 request_type=request_type)
        LOG.debug('Create snapshot %(snap)s for share %(share)s',
                  {'snap': snap_name, 'share': share_name})

    @inspur_driver_debug_trace
    def delete_snapshot(self, context, snapshot, share_server=None):
        """delete snapshot of share"""
        # !!! Attention the share property is a ShareInstance
        share = snapshot['share']
        pool, share_name, _, _ = self._get_share_instance_pnsp(share)
        share_path = self._generate_share_path(pool, share_name)

        # if there are no snapshot exist, driver will return directly
        snaps_backend = self._get_snapshots_from_share(share_path)
        if len(snaps_backend) == 0:
            return

        snap_name = self._generate_snapshot_name(snapshot)

        method = ('snapshot/directory?path=%s&snapName=%s'
                  % (share_path, snap_name))
        request_type = 'delete'
        self._rest.send_rest_api(method=method, request_type=request_type)
        LOG.debug('Delete snapshot %(snap)s of share %(share)s',
                  {'snap': snap_name, 'share': share_name})

    @staticmethod
    def transfer_rule_to_client(proto, rule):
        """transfer manila access rule to backend client"""
        access_level = rule['access_level']
        if proto == 'cifs' and access_level == 'rw':
            access_level = 'rwx'
        return dict(name=rule['access_to'],
                    type=(0 if proto == 'nfs' else 1),
                    authority=access_level)

    @inspur_driver_debug_trace
    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        """update access of share"""
        pool, share_name, _, proto = self._get_share_instance_pnsp(share)
        share_path = self._generate_share_path(pool, share_name)

        method = 'file/share/%s' % proto
        request_type = 'put'
        params = {
            'path': share_path,
            'addedClientList': [],
            'deletedClientList': [],
            'editedClientList': []
        }

        if proto == 'nfs':
            share_backend = self._get_nfs_share(share_path)
            params['pathAuthority'] = share_backend['pathAuthority']
        else:
            params['name'] = share_name

        if add_rules or delete_rules:
            to_add_clients = [self.transfer_rule_to_client(proto, rule)
                              for rule in add_rules]
            params['addedClientList'] = to_add_clients
            to_del_clients = [self.transfer_rule_to_client(proto, rule)
                              for rule in delete_rules]
            params['deletedClientList'] = to_del_clients
        else:
            access_clients = [self.transfer_rule_to_client(proto, rule)
                              for rule in access_rules]
            params['addedClientList'] = access_clients
            self._clear_access(share)

        self._rest.send_rest_api(method=method,
                                 params=params,
                                 request_type=request_type)
        LOG.debug('complete the update access work for share %s', share_name)

    @inspur_driver_debug_trace
    def _update_share_stats(self, data=None):
        """update the backend stats including driver info and pools info"""
        # Do a check of the token validity each time we update share stats,
        # do a refresh if token already expires
        time_difference = time.time() - self._token_time
        if time_difference > self.token_available_time:
            self._rest.refresh_token()
            self._token_time = time.time()
            LOG.debug('Token of Driver has been refreshed')

        data = {
            'vendor_name': self.VENDOR,
            'driver_version': self.VERSION,
            'storage_protocol': self.PROTOCOL,
            'share_backend_name':
                self.configuration.safe_get('share_backend_name'),
            'snapshot_support': True,
            'create_share_from_snapshot_support': True,
            'pools': [self._get_pool_stats(pool) for pool in self.pools]
        }

        super(AS13000ShareDriver, self)._update_share_stats(data)

    @inspur_driver_debug_trace
    def _clear_access(self, share):
        """clear all access of share"""
        pool, share_name, size, proto = self._get_share_instance_pnsp(share)
        share_path = self._generate_share_path(pool, share_name)

        method = 'file/share/%s' % proto
        request_type = 'put'
        params = {
            'path': share_path,
            'addedClientList': [],
            'deletedClientList': [],
            'editedClientList': []
        }

        if proto == 'nfs':
            share_backend = self._get_nfs_share(share_path)
            params['deletedClientList'] = share_backend['clientList']
            params['pathAuthority'] = share_backend['pathAuthority']
        else:
            share_backend = self._get_cifs_share(share_name)
            params['deletedClientList'] = share_backend['userList']
            params['name'] = share_name

        self._rest.send_rest_api(method=method,
                                 params=params,
                                 request_type=request_type)
        LOG.debug('Clear all the access of share %s', share_name)

    @inspur_driver_debug_trace
    def _validate_pools_exist(self):
        """Check the pool in conf exist in the backend"""
        available_pools = self._get_directory_list('/')
        for pool in self.pools:
            if pool not in available_pools:
                msg = (_('Pool %s is not exist in backend storage.') % pool)
                LOG.error(msg)
                raise exception.InvalidInput(reason=msg)

    @inspur_driver_debug_trace
    def _get_directory_quota(self, path):
        """get the quota of directory"""
        method = 'file/quota/directory?path=/%s' % path
        request_type = 'get'
        data = self._rest.send_rest_api(method=method,
                                        request_type=request_type)
        quota = data.get('hardthreshold')
        if quota is None:
            # the method of '_update_share_stats' will check quota of pools.
            # To avoid return NONE for pool info, so raise this exception
            msg = (_(r'Quota of pool: /%s is not set, '
                     r'please set it in GUI of AS13000') % path)
            LOG.error(msg)
            raise exception.ShareBackendException(msg=msg)

        hardunit = data.get('hardunit')
        used_capacity = data.get('capacity')
        used_capacity = (str(used_capacity)).upper()
        used_capacity = self._unit_convert(used_capacity)

        if hardunit == 1:
            quota = quota * 1024
        total_capacity = int(quota)
        used_capacity = int(used_capacity)
        return total_capacity, used_capacity

    def _get_pool_stats(self, path):
        """Get the stats of pools, such as capacity and other information."""

        total_capacity, used_capacity = self._get_directory_quota(path)
        free_capacity = total_capacity - used_capacity

        pool = {
            'pool_name': path,
            'reserved_percentage':
                self.configuration.reserved_share_percentage,
            'reserved_snapshot_percentage':
                self.configuration.reserved_share_from_snapshot_percentage
                or self.configuration.reserved_share_percentage,
            'max_over_subscription_ratio':
                self.configuration.max_over_subscription_ratio,
            'dedupe': False,
            'compression': False,
            'qos': False,
            'thin_provisioning': True,
            'total_capacity_gb': total_capacity,
            'free_capacity_gb': free_capacity,
            'allocated_capacity_gb': used_capacity,
            'snapshot_support': True,
            'create_share_from_snapshot_support': True
        }

        return pool

    @inspur_driver_debug_trace
    def _get_directory_list(self, path):
        """Get all the directory list of target path"""
        method = 'file/directory?path=%s' % path
        request_type = 'get'
        directory_list = self._rest.send_rest_api(method=method,
                                                  request_type=request_type)
        dir_list = []
        for directory in directory_list:
            dir_list.append(directory['name'])
        return dir_list

    @inspur_driver_debug_trace
    def _create_directory(self, share_name, pool_name):
        """create a directory for share"""

        method = 'file/directory'
        request_type = 'post'
        params = {'name': share_name,
                  'parentPath': self.base_dir_detail['path'],
                  'authorityInfo': self.base_dir_detail['authorityInfo'],
                  'dataProtection': self.base_dir_detail['dataProtection'],
                  'poolName': self.base_dir_detail['poolName']}
        self._rest.send_rest_api(method=method,
                                 params=params,
                                 request_type=request_type)

        return self._generate_share_path(pool_name, share_name)

    @inspur_driver_debug_trace
    def _delete_directory(self, share_path):
        """delete the directory when delete share"""
        method = 'file/directory?path=%s' % share_path
        request_type = 'delete'
        self._rest.send_rest_api(method=method, request_type=request_type)

    @inspur_driver_debug_trace
    def _set_directory_quota(self, share_path, quota):
        """set directory quota for share"""
        method = 'file/quota/directory'
        request_type = 'put'
        params = {'path': share_path, 'hardthreshold': quota, 'hardunit': 2}
        self._rest.send_rest_api(method=method,
                                 params=params,
                                 request_type=request_type)

    @inspur_driver_debug_trace
    def _create_nfs_share(self, share_path):
        """create a NFS share"""
        method = 'file/share/nfs'
        request_type = 'post'
        params = {'path': share_path, 'pathAuthority': 'rw', 'client': []}
        self._rest.send_rest_api(method=method,
                                 params=params,
                                 request_type=request_type)

    @inspur_driver_debug_trace
    def _delete_nfs_share(self, share_path):
        """Delete the NFS share"""
        method = 'file/share/nfs?path=%s' % share_path
        request_type = 'delete'
        self._rest.send_rest_api(method=method, request_type=request_type)

    @inspur_driver_debug_trace
    def _get_nfs_share(self, share_path):
        """Get the nfs share in backend"""
        method = 'file/share/nfs?path=%s' % share_path
        request_type = 'get'
        share_backend = self._rest.send_rest_api(method=method,
                                                 request_type=request_type)
        return share_backend

    @inspur_driver_debug_trace
    def _create_cifs_share(self, share_name, share_path):
        """Create a CIFS share."""
        method = 'file/share/cifs'
        request_type = 'post'
        params = {'path': share_path,
                  'name': share_name,
                  'userlist': []}
        self._rest.send_rest_api(method=method,
                                 params=params,
                                 request_type=request_type)

    @inspur_driver_debug_trace
    def _delete_cifs_share(self, share_name):
        """Delete the CIFS share."""
        method = 'file/share/cifs?name=%s' % share_name
        request_type = 'delete'
        self._rest.send_rest_api(method=method, request_type=request_type)

    @inspur_driver_debug_trace
    def _get_cifs_share(self, share_name):
        """Get the CIFS share in backend"""
        method = 'file/share/cifs?name=%s' % share_name
        request_type = 'get'
        share_backend = self._rest.send_rest_api(method=method,
                                                 request_type=request_type)
        return share_backend

    @inspur_driver_debug_trace
    def _clone_directory_to_dest(self, snapshot, dest_path):
        """Clone the directory to the new directory"""
        # get the origin share name of the snapshot
        share_instance = snapshot['share_instance']
        pool, name, _, _ = self._get_share_instance_pnsp(share_instance)
        share_path = self._generate_share_path(pool, name)

        # get the snapshot instance name
        snap_name = self._generate_snapshot_name(snapshot)

        method = 'snapshot/directory/clone'
        request_type = 'post'
        params = {'path': share_path,
                  'snapName': snap_name,
                  'destPath': dest_path}
        self._rest.send_rest_api(method=method,
                                 params=params,
                                 request_type=request_type)
        LOG.debug('Clone Path: %(path)s Snapshot: %(snap)s to Path %(dest)s',
                  {'path': share_path, 'snap': snap_name, 'dest': dest_path})

    @inspur_driver_debug_trace
    def _get_snapshots_from_share(self, path):
        """get all the snapshot of share"""
        method = 'snapshot/directory?path=%s' % path
        request_type = 'get'
        snaps = self._rest.send_rest_api(method=method,
                                         request_type=request_type)
        return snaps

    @inspur_driver_debug_trace
    def _get_location_path(self, share_name, share_path, share_proto):
        """return all the location of all nodes"""
        if share_proto == 'nfs':
            location = [
                {'path': r'%(ip)s:%(share_path)s'
                         % {'ip': ip, 'share_path': share_path}}
                for ip in self.ips]
        else:
            location = [
                {'path': r'\\%(ip)s\%(share_name)s'
                         % {'ip': ip, 'share_name': share_name}}
                for ip in self.ips]

        return location

    def _get_nodes_virtual_ips(self):
        """Get the virtual ip list of the node"""
        method = 'ctdb/set'
        request_type = 'get'
        ctdb_set = self._rest.send_rest_api(method=method,
                                            request_type=request_type)
        virtual_ips = []
        for vip in ctdb_set['virtualIpList']:
            ip = vip['ip'].split('/')[0]
            virtual_ips.append(ip)
        return virtual_ips

    def _get_nodes_physical_ips(self):
        """Get the physical ip of all the backend nodes"""
        method = 'cluster/node/cache'
        request_type = 'get'
        cached_nodes = self._rest.send_rest_api(method=method,
                                                request_type=request_type)
        node_ips = []
        for node in cached_nodes:
            if node['runningStatus'] == 1 and node['healthStatus'] == 1:
                node_ips.append(node['nodeIp'])

        return node_ips

    def _get_nodes_ips(self):
        """Return both the physical ip and virtual ip"""
        virtual_ips = self._get_nodes_virtual_ips()
        physical_ips = self._get_nodes_physical_ips()

        return virtual_ips + physical_ips

    def _get_share_instance_pnsp(self, share_instance):
        """Get pool, name, size, proto information of a share instance.

        AS13000 require all the names can only consist of letters,numbers,
        and undercores,and must begin with a letter.
        Also the length of name must less than 32 character.
        The driver will use the ID as the name in backend,
        add 'share_' to the beginning,and convert '-' to '_'
        """
        pool = share_utils.extract_host(share_instance['host'], level='pool')
        name = self._generate_share_name(share_instance)
        # a share instance may not contain size attr.
        try:
            size = share_instance['size']
        except AttributeError:
            size = None

        # a share instance may not contain proto attr.
        try:
            proto = share_instance['share_proto'].lower()
        except AttributeError:
            proto = None

        LOG.debug("Pool %s, Name: %s, Size: %s, Protocol: %s",
                  pool, name, size, proto)

        return pool, name, size, proto

    def _unit_convert(self, capacity):
        """Convert all units to GB"""
        capacity = str(capacity)
        capacity = capacity.upper()
        try:
            unit_of_used = re.findall(r'[A-Z]', capacity)
            unit_of_used = ''.join(unit_of_used)
        except BaseException:
            unit_of_used = ''
        capacity = capacity.replace(unit_of_used, '')
        capacity = float(capacity.replace(unit_of_used, ''))
        if unit_of_used in ['B', '']:
            capacity = capacity / units.Gi
        elif unit_of_used in ['K', 'KB']:
            capacity = capacity / units.Mi
        elif unit_of_used in ['M', 'MB']:
            capacity = capacity / units.Ki
        elif unit_of_used in ['G', 'GB']:
            capacity = capacity
        elif unit_of_used in ['T', 'TB']:
            capacity = capacity * units.Ki
        elif unit_of_used in ['E', 'EB']:
            capacity = capacity * units.Mi

        capacity = '%.0f' % capacity
        return float(capacity)

    def _format_name(self, name):
        """format name to meet the backend requirements"""
        name = name[0:32]
        name = name.replace('-', '_')
        return name

    def _generate_share_name(self, share_instance):
        share_name = 'share_%s' % share_instance['id']
        return self._format_name(share_name)

    def _generate_snapshot_name(self, snapshot_instance):
        snap_name = 'snap_%s' % snapshot_instance['id']
        return self._format_name(snap_name)

    @staticmethod
    def _generate_share_path(pool, share_name):
        return r'/%s/%s' % (pool, share_name)

    def _get_directory_detail(self, directory):
        method = 'file/directory/detail?path=/%s' % directory
        request_type = 'get'
        details = self._rest.send_rest_api(method=method,
                                           request_type=request_type)
        return details[0]
